"""Collects commonly-used core classes & functions for easier import.

Ex: `from nems import Model`

"""

from nems.models.base import Model
from nems.tools.demo_data.file_management import download_demo, load_demo
